! WHIZARD 2.0.0 Mon Aug 24 2009
! 
! (C) 1999-2009 by 
!     Wolfgang Kilian <kilian@hep.physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@physik.uni-freiburg.de>
!     with contributions by Sebastian Schmidt
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module pdg_arrays

  use file_utils

  implicit none
  private

  public :: pdg_array_t
  public :: pdg_array_write
  public :: assignment(=)
  public :: operator(//)
  public :: operator(.match.)
  public :: operator(.eqv.)
  public :: operator(.neqv.)

  integer, parameter, public :: UNDEFINED = 0

  type :: pdg_array_t
     private
     integer, dimension(:), allocatable :: pdg
  end type pdg_array_t


  interface assignment(=)
     module procedure pdg_array_from_int_array
     module procedure pdg_array_from_int
     module procedure int_array_from_pdg_array
  end interface

  interface operator(//)
     module procedure concat_pdg_arrays
  end interface

  interface operator(.match.)
     module procedure pdg_array_match_integer
  end interface

  interface operator(.eqv.)
     module procedure pdg_array_equivalent
  end interface
  interface operator(.neqv.)
     module procedure pdg_array_inequivalent
  end interface


contains

  subroutine pdg_array_write (cval, unit)
    type(pdg_array_t), intent(in) :: cval
    integer, intent(in), optional :: unit
    integer :: u, i
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(A)", advance="no")  "PDG("
    if (allocated (cval%pdg)) then
       do i = 1, size (cval%pdg)
          if (i > 1)  write (u, "(A)", advance="no")  ", "
          write (u, "(I0)", advance="no")  cval%pdg(i)
       end do
    end if
    write (u, "(A)", advance="no")  ")"
  end subroutine pdg_array_write

  subroutine pdg_array_from_int_array (cval, iarray)
    type(pdg_array_t), intent(out) :: cval
    integer, dimension(:), intent(in) :: iarray
    allocate (cval%pdg (size (iarray)))
    cval%pdg = iarray
  end subroutine pdg_array_from_int_array

  elemental subroutine pdg_array_from_int (cval, int)
    type(pdg_array_t), intent(out) :: cval
    integer, intent(in) :: int
    allocate (cval%pdg (1))
    cval%pdg = int
  end subroutine pdg_array_from_int

  subroutine int_array_from_pdg_array (iarray, cval)
    integer, dimension(:), allocatable, intent(out) :: iarray
    type(pdg_array_t), intent(in) :: cval
    allocate (iarray (size (cval%pdg)))
    iarray = cval%pdg
  end subroutine int_array_from_pdg_array

  function concat_pdg_arrays (cval1, cval2) result (cval)
    type(pdg_array_t) :: cval
    type(pdg_array_t), intent(in) :: cval1, cval2
    integer :: n1, n2
    n1 = size (cval1%pdg)
    n2 = size (cval2%pdg)
    allocate (cval%pdg (n1 + n2))
    cval%pdg(:n1) = cval1%pdg
    cval%pdg(n1+1:) = cval2%pdg
  end function concat_pdg_arrays

  elemental function pdg_array_match_integer (cval, pdg) result (flag)
    logical :: flag
    type(pdg_array_t), intent(in) :: cval
    integer, intent(in) :: pdg
    flag = pdg == UNDEFINED &
         .or. any (cval%pdg == UNDEFINED) &
         .or. any (cval%pdg == pdg)
  end function pdg_array_match_integer

  function pdg_array_equivalent (cval1, cval2) result (eq)
    logical :: eq
    type(pdg_array_t), intent(in) :: cval1, cval2
    logical, dimension(:), allocatable :: match1, match2
    integer :: i
    eq = any (cval1%pdg == UNDEFINED) &
         .and. any (cval1%pdg == UNDEFINED)
    if (.not. eq) then
       allocate (match1 (size (cval1%pdg)))
       allocate (match2 (size (cval2%pdg)))
       match1 = .false.
       match2 = .false.
       do i = 1, size (cval1%pdg)
          match2 = match2 .or. cval1%pdg(i) == cval2%pdg
       end do
       do i = 1, size (cval2%pdg)
          match1 = match1 .or. cval2%pdg(i) == cval1%pdg
       end do
       eq = all (match1) .and. all (match2)
    end if
  end function pdg_array_equivalent

  function pdg_array_inequivalent (cval1, cval2) result (neq)
    logical :: neq
    type(pdg_array_t), intent(in) :: cval1, cval2
    neq = .not. pdg_array_equivalent (cval1, cval2)
  end function pdg_array_inequivalent


end module pdg_arrays
